@@ -1,2 +1,7 @@
 from .EndianBinaryReader import EndianBinaryReader
 from .EndianBinaryWriter import EndianBinaryWriter
+
+__all__ = [
+    "EndianBinaryReader",
+    "EndianBinaryWriter",
+]
